#include <JTC/JTC.h>
#include <iostream>

using namespace std;

enum DIREZIONE {MARE=0, MONTAGNA=1};

class Ponte : public JTCMonitor
{
	DIREZIONE direzione_corrente;
    int turisti_sul_ponte;

public:
	Ponte()
	{
		direzione_corrente = MARE;
		turisti_sul_ponte = 0;
	}
	void iniziaAttraversamento(DIREZIONE direzione)
	{
		JTCSynchronized sync(*this);
		while(direzione != direzione_corrente && turisti_sul_ponte != 0)
			wait();
		direzione_corrente = direzione;
		++turisti_sul_ponte;
	}
	void fineAttraversamento()
	{
		JTCSynchronized sync(*this);
		--turisti_sul_ponte;
		if (turisti_sul_ponte == 0)
			notifyAll();
	}
};

class PonteNS : public JTCMonitor
/*
 * Ponte senza starvation
 */
{
	DIREZIONE direzione_corrente;
    int in_attesa_per[2];
    int turisti_sul_ponte;
    int numero_cortesia;
    int turisti_stessa_direzione;

    DIREZIONE opposta(DIREZIONE direzione)
    {
        return (direzione == MARE ? MONTAGNA : MARE);
    }

public:
	PonteNS()
	{
		direzione_corrente = MARE;
        in_attesa_per[MARE] = 0;
        in_attesa_per[MONTAGNA] = 0;
		turisti_sul_ponte = 0;
        numero_cortesia = 3;
        turisti_stessa_direzione = 0;
	}
	void iniziaAttraversamento(DIREZIONE direzione)
	{
		JTCSynchronized sync(*this);
        ++in_attesa_per[direzione];		
        while ( (direzione != direzione_corrente && 
                 turisti_sul_ponte != 0) ||
                (direzione == direzione_corrente && 
                 turisti_stessa_direzione >= numero_cortesia &&
                 in_attesa_per[opposta(direzione)] > 0) )
        {
			wait();
        }
        --in_attesa_per[direzione];
        if ( direzione == direzione_corrente )
            ++turisti_stessa_direzione;
        else
        {
		    direzione_corrente = direzione;
            turisti_stessa_direzione = 1;
        }
		++turisti_sul_ponte;
	}
	void fineAttraversamento()
	{
		JTCSynchronized sync(*this);
		--turisti_sul_ponte;
		if (turisti_sul_ponte == 0)
			notifyAll();
	}
};

Ponte ponte;

class Turista : public JTCThread
{
	int id;
	DIREZIONE direzione;

public:
	Turista(int i, DIREZIONE d)
	{
		id = i;
		direzione = d;
	}
	virtual void run()
	{
		while(true) 
        {
			cout << "Il Turista " << id 
                 << " e' in attesa di attraversare il ponte nella direzione " 
                 << (direzione == MARE ? "MARE." : "MONTAGNA.") << endl;
			ponte.iniziaAttraversamento(direzione);
			
            cout << "Il Turista " << id 
                 << " sta attraversando il ponte." << endl;
			sleep(rand() % 1000);
			
            cout << "Il Turista " << id 
                 << " ha attraversato il ponte." << endl;
			ponte.fineAttraversamento();
			
            sleep(rand() % 1000);
			//direzione = (direzione == MARE ? MONTAGNA : MARE);
		}
	}
};


int main(int argc, char* argv[])
{   
	JTCInitialize initialize;
	Turista* t[10];
    for (int i = 0; i < 9; i++) 
    {
		t[i] = new Turista(i,MARE);
		t[i]->start();
	}
    for (int i = 9; i < 10; i++) 
    {
		t[i] = new Turista(i, MONTAGNA);
		t[i]->start();
	}
	return 0;	
}

